@@ -24,11 +24,13 @@
 #include "goption.h"
 #include "glib.h"
 #include "glibintl.h"
+#include "gprintf.h"
 
 #include "galias.h"
 
 #include <string.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <errno.h>
 
 #define TRANSLATE(group, str) (((group)->translate_func ? (* (group)->translate_func) ((str), (group)->translate_data) : (str)))
@@ -50,6 +52,8 @@ typedef struct
     gint integer;
     gchar *str;
     gchar **array;
+    gdouble dbl;
+    gint64 int64;
   } prev;
   union 
   {
@@ -70,39 +74,45 @@ typedef struct
 
 struct _GOptionContext
 {
-  GList *groups;
+  GList           *groups;
 
-  gchar *parameter_string;
+  gchar           *parameter_string;
+  gchar           *summary;
+  gchar           *description;
 
-  gboolean help_enabled;
-  gboolean ignore_unknown;
+  GTranslateFunc   translate_func;
+  GDestroyNotify   translate_notify;
+  gpointer	   translate_data;
+
+  guint            help_enabled   : 1;
+  guint            ignore_unknown : 1;
   
-  GOptionGroup *main_group;
+  GOptionGroup    *main_group;
 
   /* We keep a list of change so we can revert them */
-  GList *changes;
+  GList           *changes;
   
-  /* We also keep track of all argv elements that should be NULLed or
-   * modified.
+  /* We also keep track of all argv elements 
+   * that should be NULLed or modified.
    */
-  GList *pending_nulls;
+  GList           *pending_nulls;
 };
 
 struct _GOptionGroup
 {
-  gchar *name;
-  gchar *description;
-  gchar *help_description;
+  gchar           *name;
+  gchar           *description;
+  gchar           *help_description;
 
-  GDestroyNotify  destroy_notify;
-  gpointer        user_data;
+  GDestroyNotify   destroy_notify;
+  gpointer         user_data;
 
-  GTranslateFunc  translate_func;
-  GDestroyNotify  translate_notify;
-  gpointer	  translate_data;
+  GTranslateFunc   translate_func;
+  GDestroyNotify   translate_notify;
+  gpointer	   translate_data;
 
-  GOptionEntry *entries;
-  gint         n_entries;
+  GOptionEntry    *entries;
+  gint             n_entries;
 
   GOptionParseFunc pre_parse_func;
   GOptionParseFunc post_parse_func;
@@ -114,25 +124,90 @@ static void free_changes_list (GOptionContext *context,
 static void free_pending_nulls (GOptionContext *context,
 				gboolean        perform_nulls);
 
+
+static int
+_g_unichar_get_width (gunichar c)
+{
+  if (G_UNLIKELY (g_unichar_iszerowidth (c)))
+    return 0;
+
+  /* we ignore the fact that we should call g_unichar_iswide_cjk() under
+   * some locales (legacy East Asian ones) */
+  if (g_unichar_iswide (c))
+    return 2;
+
+  return 1;
+}
+
+static glong
+_g_utf8_strwidth (const gchar *p,
+                  gssize       max)
+{
+  glong len = 0;
+  const gchar *start = p;
+  g_return_val_if_fail (p != NULL || max == 0, 0);
+
+  if (max < 0)
+    {
+      while (*p)
+        {
+          len += _g_unichar_get_width (g_utf8_get_char (p));
+          p = g_utf8_next_char (p);
+        }
+    }
+  else
+    {
+      if (max == 0 || !*p)
+        return 0;
+
+      /* this case may not be quite correct */
+      
+      len += _g_unichar_get_width (g_utf8_get_char (p));
+      p = g_utf8_next_char (p);          
+
+      while (p - start < max && *p)
+        {
+          len += _g_unichar_get_width (g_utf8_get_char (p));
+          p = g_utf8_next_char (p);          
+        }
+    }
+
+  return len;
+}
+
+
 GQuark
 g_option_error_quark (void)
 {
-  static GQuark q = 0;
-  
-  if (q == 0)
-    q = g_quark_from_static_string ("g-option-context-error-quark");
-
-  return q;
+  return g_quark_from_static_string ("g-option-context-error-quark");
 }
 
 /**
  * g_option_context_new:
  * @parameter_string: a string which is displayed in
- *    the first line of <option>--help</option> output, after 
+ *    the first line of <option>--help</option> output, after the
+ *    usage summary 
  *    <literal><replaceable>programname</replaceable> [OPTION...]</literal>
  *
  * Creates a new option context. 
  *
+ * The @parameter_string can serve multiple purposes. It can be used
+ * to add descriptions for "rest" arguments, which are not parsed by
+ * the #GOptionContext, typically something like "FILES" or
+ * "FILE1 FILE2...". If you are using #G_OPTION_REMAINING for
+ * collecting "rest" arguments, GLib handles this automatically by
+ * using the @arg_description of the corresponding #GOptionEntry in
+ * the usage summary.
+ *
+ * Another usage is to give a short summary of the program
+ * functionality, like " - frob the strings", which will be displayed
+ * in the same line as the usage. For a longer description of the
+ * program functionality that should be displayed as a paragraph
+ * below the usage line, use g_option_context_set_summary().
+ *
+ * Note that the @parameter_string is translated (see
+ * g_option_context_set_translate_func()). 
+ *
  * Returns: a newly created #GOptionContext, which must be
  *    freed with g_option_context_free() after use.
  *
@@ -176,7 +251,12 @@ void g_option_context_free (GOptionContext *context)
   free_pending_nulls (context, FALSE);
   
   g_free (context->parameter_string);
+  g_free (context->summary);
+  g_free (context->description);
   
+  if (context->translate_notify)
+    (* context->translate_notify) (context->translate_data);
+
   g_free (context);
 }
 
@@ -395,13 +475,13 @@ calculate_max_length (GOptionGroup *group)
       if (entry->flags & G_OPTION_FLAG_HIDDEN)
 	continue;
 
-      len = g_utf8_strlen (entry->long_name, -1);
+      len = _g_utf8_strwidth (entry->long_name, -1);
       
       if (entry->short_name)
 	len += 4;
       
       if (!NO_ARG (entry) && entry->arg_description)
-	len += 1 + g_utf8_strlen (TRANSLATE (group, entry->arg_description), -1);
+	len += 1 + _g_utf8_strwidth (TRANSLATE (group, entry->arg_description), -1);
       
       max_length = MAX (max_length, len);
     }
@@ -412,10 +492,11 @@ calculate_max_length (GOptionGroup *group)
 static void
 print_entry (GOptionGroup       *group,
 	     gint                max_length,
-	     const GOptionEntry *entry)
+	     const GOptionEntry *entry,
+             GString            *string)
 {
   GString *str;
-
+  
   if (entry->flags & G_OPTION_FLAG_HIDDEN)
     return;
 
@@ -432,15 +513,34 @@ print_entry (GOptionGroup       *group,
   if (entry->arg_description)
     g_string_append_printf (str, "=%s", TRANSLATE (group, entry->arg_description));
   
-  g_print ("%-*s %s\n", max_length + 4, str->str,
-	   entry->description ? TRANSLATE (group, entry->description) : "");
+  g_string_append_printf (string, "%s%*s %s\n", str->str,
+	                  (int) (max_length + 4 - _g_utf8_strwidth (str->str, -1)), "",
+	                  entry->description ? TRANSLATE (group, entry->description) : "");
   g_string_free (str, TRUE);  
 }
 
-static void
-print_help (GOptionContext *context,
-	    gboolean        main_help,
-	    GOptionGroup   *group) 
+/**
+ * g_option_context_get_help: 
+ * @context: a #GOptionContext
+ * @main_help: if %TRUE, only include the main group 
+ * @group: the #GOptionGroup to create help for, or %NULL
+ *
+ * Returns a formatted, translated help text for the given context.
+ * To obtain the text produced by <option>--help</option>, call
+ * <literal>g_option_context_get_help (context, TRUE, NULL)</literal>.
+ * To obtain the text produced by <option>--help-all</option>, call
+ * <literal>g_option_context_get_help (context, FALSE, NULL)</literal>.
+ * To obtain the help text for an option group, call
+ * <literal>g_option_context_get_help (context, FALSE, group)</literal>.
+ *
+ * Returns: A newly allocated string containing the help text
+ *
+ * Since: 2.14
+ */
+gchar *
+g_option_context_get_help (GOptionContext *context,
+	                   gboolean        main_help,
+	                   GOptionGroup   *group)
 {
   GList *list;
   gint max_length, len;
@@ -449,27 +549,47 @@ print_help (GOptionContext *context,
   GHashTable *shadow_map;
   gboolean seen[256];
   const gchar *rest_description;
-  
+  GString *string;
+
+  string = g_string_sized_new (1024);
+
   rest_description = NULL;
   if (context->main_group)
     {
+
       for (i = 0; i < context->main_group->n_entries; i++)
 	{
 	  entry = &context->main_group->entries[i];
 	  if (entry->long_name[0] == 0)
 	    {
-	      rest_description = entry->arg_description;
+	      rest_description = TRANSLATE (context->main_group, entry->arg_description);
 	      break;
 	    }
 	}
     }
-  
-  g_print ("%s\n  %s %s%s%s%s%s\n\n", 
-	   _("Usage:"), g_get_prgname(), _("[OPTION...]"),
-	   rest_description ? " " : "",
-	   rest_description ? rest_description : "",
-	   context->parameter_string ? " " : "",
-	   context->parameter_string ? context->parameter_string : "");
+
+  g_string_append_printf (string, "%s\n  %s %s", 
+	                  _("Usage:"), g_get_prgname(), _("[OPTION...]"));
+
+  if (rest_description)
+    {
+      g_string_append (string, " ");
+      g_string_append (string, rest_description);
+    }
+
+  if (context->parameter_string)
+    {
+      g_string_append (string, " ");
+      g_string_append (string, TRANSLATE (context, context->parameter_string));
+    }
+
+  g_string_append (string, "\n\n");
+
+  if (context->summary)
+    {
+      g_string_append (string, TRANSLATE (context, context->summary));
+      g_string_append (string, "\n\n");
+    }
 
   memset (seen, 0, sizeof (gboolean) * 256);
   shadow_map = g_hash_table_new (g_str_hash, g_str_equal);
@@ -498,13 +618,13 @@ print_help (GOptionContext *context,
 	{
 	  entry = &group->entries[i];
 	  if (g_hash_table_lookup (shadow_map, entry->long_name) && 
-	      !(entry->flags && G_OPTION_FLAG_NOALIAS))
+	      !(entry->flags & G_OPTION_FLAG_NOALIAS))
 	    entry->long_name = g_strdup_printf ("%s-%s", group->name, entry->long_name);
 	  else  
 	    g_hash_table_insert (shadow_map, (gpointer)entry->long_name, entry);
 
 	  if (seen[(guchar)entry->short_name] && 
-	      !(entry->flags && G_OPTION_FLAG_NOALIAS))
+	      !(entry->flags & G_OPTION_FLAG_NOALIAS))
 	    entry->short_name = 0;
 	  else
 	    seen[(guchar)entry->short_name] = TRUE;
@@ -516,11 +636,11 @@ print_help (GOptionContext *context,
 
   list = context->groups;
 
-  max_length = g_utf8_strlen ("-?, --help", -1);
+  max_length = _g_utf8_strwidth ("-?, --help", -1);
 
   if (list)
     {
-      len = g_utf8_strlen ("--help-all", -1);
+      len = _g_utf8_strwidth ("--help-all", -1);
       max_length = MAX (max_length, len);
     }
 
@@ -535,7 +655,7 @@ print_help (GOptionContext *context,
       GOptionGroup *group = list->data;
       
       /* First, we check the --help-<groupname> options */
-      len = g_utf8_strlen ("--help-", -1) + g_utf8_strlen (group->name, -1);
+      len = _g_utf8_strwidth ("--help-", -1) + _g_utf8_strwidth (group->name, -1);
       max_length = MAX (max_length, len);
 
       /* Then we go through the entries */
@@ -552,36 +672,39 @@ print_help (GOptionContext *context,
     {
       list = context->groups;
       
-      g_print ("%s\n  -%c, --%-*s %s\n", 
-	       _("Help Options:"), '?', max_length - 4, "help", 
-	       _("Show help options"));
+      g_string_append_printf (string, "%s\n  -%c, --%-*s %s\n", 
+	                      _("Help Options:"), '?', max_length - 4, "help", 
+	                      _("Show help options"));
       
       /* We only want --help-all when there are groups */
       if (list)
-	g_print ("  --%-*s %s\n", max_length, "help-all", 
-		 _("Show all help options"));
+	g_string_append_printf (string, "  --%-*s %s\n", 
+                                max_length, "help-all", 
+                                _("Show all help options"));
       
       while (list)
 	{
 	  GOptionGroup *group = list->data;
 	  
-	  g_print ("  --help-%-*s %s\n", max_length - 5, group->name, 
-		   TRANSLATE (group, group->help_description));
+	  g_string_append_printf (string, "  --help-%-*s %s\n", 
+                                  max_length - 5, group->name, 
+		                  TRANSLATE (group, group->help_description));
 	  
 	  list = list->next;
 	}
 
-      g_print ("\n");
+      g_string_append (string, "\n");
     }
 
   if (group)
     {
       /* Print a certain group */
       
-      g_print ("%s\n", TRANSLATE (group, group->description));
+      g_string_append (string, TRANSLATE (group, group->description));
+      g_string_append (string, "\n");
       for (i = 0; i < group->n_entries; i++)
-	print_entry (group, max_length, &group->entries[i]);
-      g_print ("\n");
+	print_entry (group, max_length, &group->entries[i], string);
+      g_string_append (string, "\n");
     }
   else if (!main_help)
     {
@@ -593,13 +716,13 @@ print_help (GOptionContext *context,
 	{
 	  GOptionGroup *group = list->data;
 
-	  g_print ("%s\n", group->description);
-
+	  g_string_append (string, group->description);
+          g_string_append (string, "\n");
 	  for (i = 0; i < group->n_entries; i++)
 	    if (!(group->entries[i].flags & G_OPTION_FLAG_IN_MAIN))
-	      print_entry (group, max_length, &group->entries[i]);
+	      print_entry (group, max_length, &group->entries[i], string);
 	  
-	  g_print ("\n");
+	  g_string_append (string, "\n");
 	  list = list->next;
 	}
     }
@@ -609,12 +732,12 @@ print_help (GOptionContext *context,
     {
       list = context->groups;
 
-      g_print ("%s\n", _("Application Options:"));
-
+      g_string_append (string,  _("Application Options:"));
+      g_string_append (string, "\n");
       if (context->main_group)
 	for (i = 0; i < context->main_group->n_entries; i++) 
 	  print_entry (context->main_group, max_length, 
-		       &context->main_group->entries[i]);
+		       &context->main_group->entries[i], string);
 
       while (list != NULL)
 	{
@@ -623,15 +746,35 @@ print_help (GOptionContext *context,
 	  /* Print main entries from other groups */
 	  for (i = 0; i < group->n_entries; i++)
 	    if (group->entries[i].flags & G_OPTION_FLAG_IN_MAIN)
-	      print_entry (group, max_length, &group->entries[i]);
+	      print_entry (group, max_length, &group->entries[i], string);
 	  
 	  list = list->next;
 	}
 
-      g_print ("\n");
+      g_string_append (string, "\n");
     }
-  
-  exit (0);
+ 
+  if (context->description)
+    {
+      g_string_append (string, TRANSLATE (context, context->description));
+      g_string_append (string, "\n");
+    }
+
+  return g_string_free (string, FALSE);
+}
+
+static void
+print_help (GOptionContext *context,
+	    gboolean        main_help,
+	    GOptionGroup   *group)
+{
+  gchar *help;
+
+  help = g_option_context_get_help (context, main_help, group);
+  g_printf (help);
+  g_free (help);
+
+  exit (0);  
 }
 
 static gboolean
@@ -668,6 +811,77 @@ parse_int (const gchar *arg_name,
   return TRUE;
 }
 
+
+static gboolean
+parse_double (const gchar *arg_name,
+	   const gchar *arg,
+	   gdouble        *result,
+	   GError     **error)
+{
+  gchar *end;
+  gdouble tmp;
+
+  errno = 0;
+  tmp = g_strtod (arg, &end);
+  
+  if (*arg == '\0' || *end != '\0')
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Cannot parse double value '%s' for %s"),
+		   arg, arg_name);
+      return FALSE;
+    }
+  if (errno == ERANGE)
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Double value '%s' for %s out of range"),
+		   arg, arg_name);
+      return FALSE;
+    }
+
+  *result = tmp;
+  
+  return TRUE;
+}
+
+
+static gboolean
+parse_int64 (const gchar *arg_name,
+	     const gchar *arg,
+	     gint64      *result,
+	     GError     **error)
+{
+  gchar *end;
+  gint64 tmp;
+
+  errno = 0;
+  tmp = g_ascii_strtoll (arg, &end, 0);
+
+  if (*arg == '\0' || *end != '\0')
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Cannot parse integer value '%s' for %s"),
+		   arg, arg_name);
+      return FALSE;
+    }
+  if (errno == ERANGE)
+    {
+      g_set_error (error,
+		   G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+		   _("Integer value '%s' for %s out of range"),
+		   arg, arg_name);
+      return FALSE;
+    }
+
+  *result = tmp;
+  
+  return TRUE;
+}
+
+
 static Change *
 get_change (GOptionContext *context,
 	    GOptionArg      arg_type,
@@ -719,7 +933,9 @@ parse_arg (GOptionContext *context,
      
 {
   Change *change;
-  
+
+  g_assert (value || OPTIONAL_ARG (entry) || NO_ARG (entry));
+
   switch (entry->arg)
     {
     case G_OPTION_ARG_NONE:
@@ -884,6 +1100,40 @@ parse_arg (GOptionContext *context,
 	
 	return retval;
 	
+	break;
+      }
+    case G_OPTION_ARG_DOUBLE:
+      {
+	gdouble data;
+
+	if (!parse_double (option_name, value,
+ 			&data,
+			error))
+	  {
+	    return FALSE;
+	  }
+
+	change = get_change (context, G_OPTION_ARG_DOUBLE,
+			     entry->arg_data);
+	change->prev.dbl = *(gdouble *)entry->arg_data;
+	*(gdouble *)entry->arg_data = data;
+	break;
+      }
+    case G_OPTION_ARG_INT64:
+      {
+        gint64 data;
+
+	if (!parse_int64 (option_name, value,
+ 			 &data,
+			 error))
+	  {
+	    return FALSE;
+	  }
+
+	change = get_change (context, G_OPTION_ARG_INT64,
+			     entry->arg_data);
+	change->prev.int64 = *(gint64 *)entry->arg_data;
+	*(gint64 *)entry->arg_data = data;
 	break;
       }
     default:
@@ -921,13 +1171,13 @@ parse_short_option (GOptionContext *context,
 	    {
 	      if (*new_index > index)
 		{
-		  g_warning ("FIXME: figure out the correct error here");
-
+		  g_set_error (error, 
+			       G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
+			       _("Error parsing option %s"), option_name);
+		  g_free (option_name);
 		  return FALSE;
 		}
 
-	      option_name = g_strdup_printf ("-%c", group->entries[j].short_name);
-	      
 	      if (index < *argc - 1)
 		{
 		  if (!OPTIONAL_ARG (&group->entries[j]))	
@@ -1105,7 +1355,8 @@ parse_remaining_arg (GOptionContext *context,
       if (group->entries[j].long_name[0])
 	continue;
 
-      g_return_val_if_fail (group->entries[j].arg == G_OPTION_ARG_STRING_ARRAY ||
+      g_return_val_if_fail (group->entries[j].arg == G_OPTION_ARG_CALLBACK ||
+                            group->entries[j].arg == G_OPTION_ARG_STRING_ARRAY ||
 			    group->entries[j].arg == G_OPTION_ARG_FILENAME_ARRAY, FALSE);
       
       add_pending_null (context, &((*argv)[*index]), NULL);
@@ -1150,6 +1401,12 @@ free_changes_list (GOptionContext *context,
 	      g_strfreev (change->allocated.array.data);
 	      *(gchar ***)change->arg_data = change->prev.array;
 	      break;
+	    case G_OPTION_ARG_DOUBLE:
+	      *(gdouble *)change->arg_data = change->prev.dbl;
+	      break;
+	    case G_OPTION_ARG_INT64:
+	      *(gint64 *)change->arg_data = change->prev.int64;
+	      break;
 	    default:
 	      g_assert_not_reached ();
 	    }
@@ -1195,8 +1452,8 @@ free_pending_nulls (GOptionContext *context,
 /**
  * g_option_context_parse:
  * @context: a #GOptionContext
- * @argc: a pointer to the number of command line arguments.
- * @argv: a pointer to the array of command line arguments.
+ * @argc: a pointer to the number of command line arguments
+ * @argv: a pointer to the array of command line arguments
  * @error: a return location for errors 
  * 
  * Parses the command line arguments, recognizing options
@@ -1216,6 +1473,11 @@ free_pending_nulls (GOptionContext *context,
  * @argv array contains one of the recognized help options,
  * this function will produce help output to stdout and
  * call <literal>exit (0)</literal>.
+ *
+ * Note that function depends on the 
+ * <link linkend="setlocale">current locale</link> for 
+ * automatic character set conversion of string and filename
+ * arguments.
  * 
  * Return value: %TRUE if the parsing was successful, 
  *               %FALSE if an error occurred
@@ -1232,19 +1494,20 @@ g_option_context_parse (GOptionContext   *context,
   GList *list;
 
   /* Set program name */
-  if (argc && argv && *argc)
+  if (!g_get_prgname())
     {
-      gchar *prgname;
-      
-      prgname = g_path_get_basename ((*argv)[0]);
-      g_set_prgname (prgname);
-      g_free (prgname);
+      if (argc && argv && *argc)
+	{
+	  gchar *prgname;
+	  
+	  prgname = g_path_get_basename ((*argv)[0]);
+	  g_set_prgname (prgname);
+	  g_free (prgname);
+	}
+      else
+	g_set_prgname ("<unknown>");
     }
-  else
-    {
-      g_set_prgname ("<unknown>");
-    }
-  
+
   /* Call pre-parse hooks */
   list = context->groups;
   while (list)
@@ -1375,36 +1638,23 @@ g_option_context_parse (GOptionContext   *context,
 		    continue;
 		}
 	      else
-		{
-		  /* short option */
-
-		  gint new_i, j;
+		{ /* short option */
+		  gint j, new_i = i, arg_length;
 		  gboolean *nulled_out = NULL;
-		  
 		  arg = (*argv)[i] + 1;
-
-		  new_i = i;
-
-		  if (context->ignore_unknown)
-		    nulled_out = g_new0 (gboolean, strlen (arg));
-		  
-		  for (j = 0; j < strlen (arg); j++)
+                  arg_length = strlen (arg);
+                  nulled_out = g_newa (gboolean, arg_length);
+                  memset (nulled_out, 0, arg_length * sizeof (gboolean));
+		  for (j = 0; j < arg_length; j++)
 		    {
 		      if (context->help_enabled && arg[j] == '?')
 			print_help (context, TRUE, NULL);
-		      
 		      parsed = FALSE;
-		      
 		      if (context->main_group &&
 			  !parse_short_option (context, context->main_group,
 					       i, &new_i, arg[j],
 					       argc, argv, error, &parsed))
-			{
-
-			  g_free (nulled_out);
-			  goto fail;
-			}
-
+                        goto fail;
 		      if (!parsed)
 			{
 			  /* Try the groups */
@@ -1412,47 +1662,38 @@ g_option_context_parse (GOptionContext   *context,
 			  while (list)
 			    {
 			      GOptionGroup *group = list->data;
-			      
 			      if (!parse_short_option (context, group, i, &new_i, arg[j],
 						       argc, argv, error, &parsed))
 				goto fail;
-			      
 			      if (parsed)
 				break;
-			  
 			      list = list->next;
 			    }
 			}
-
-		      if (context->ignore_unknown)
-			{
-			  if (parsed)
-			    nulled_out[j] = TRUE;
-			  else
-			    continue;
-			}
-
-		      if (!parsed)
-			break;
+                      
+		      if (context->ignore_unknown && parsed)
+                        nulled_out[j] = TRUE;
+                      else if (context->ignore_unknown)
+                        continue;
+                      else if (!parsed)
+                        break;
+                      /* !context->ignore_unknown && parsed */
 		    }
-
 		  if (context->ignore_unknown)
 		    {
 		      gchar *new_arg = NULL; 
 		      gint arg_index = 0;
-		      
-		      for (j = 0; j < strlen (arg); j++)
+		      for (j = 0; j < arg_length; j++)
 			{
 			  if (!nulled_out[j])
 			    {
 			      if (!new_arg)
-				new_arg = g_malloc (strlen (arg) + 1);
+				new_arg = g_malloc (arg_length + 1);
 			      new_arg[arg_index++] = arg[j];
 			    }
 			}
 		      if (new_arg)
 			new_arg[arg_index] = '\0';
-		      
 		      add_pending_null (context, &((*argv)[i]), new_arg);
 		    }
 		  else if (parsed)
@@ -1789,5 +2030,153 @@ g_option_group_set_translation_domain (GOptionGroup *group,
 				     g_free);
 } 
 
+/**
+ * g_option_context_set_translate_func:
+ * @context: a #GOptionContext
+ * @func: the #GTranslateFunc, or %NULL 
+ * @data: user data to pass to @func, or %NULL
+ * @destroy_notify: a function which gets called to free @data, or %NULL
+ * 
+ * Sets the function which is used to translate the contexts 
+ * user-visible strings, for <option>--help</option> output. 
+ * If @func is %NULL, strings are not translated.
+ *
+ * Note that option groups have their own translation functions, 
+ * this function only affects the @parameter_string (see g_option_context_new()), 
+ * the summary (see g_option_context_set_summary()) and the description 
+ * (see g_option_context_set_description()).
+ *
+ * If you are using gettext(), you only need to set the translation
+ * domain, see g_context_group_set_translation_domain().
+ *
+ * Since: 2.12
+ **/
+void
+g_option_context_set_translate_func (GOptionContext *context,
+				     GTranslateFunc func,
+				     gpointer       data,
+				     GDestroyNotify destroy_notify)
+{
+  g_return_if_fail (context != NULL);
+  
+  if (context->translate_notify)
+    context->translate_notify (context->translate_data);
+      
+  context->translate_func = func;
+  context->translate_data = data;
+  context->translate_notify = destroy_notify;
+}
+
+/**
+ * g_option_context_set_translation_domain:
+ * @context: a #GOptionContext
+ * @domain: the domain to use
+ * 
+ * A convenience function to use gettext() for translating
+ * user-visible strings. 
+ * 
+ * Since: 2.12
+ **/
+void
+g_option_context_set_translation_domain (GOptionContext *context,
+					 const gchar     *domain)
+{
+  g_return_if_fail (context != NULL);
+
+  g_option_context_set_translate_func (context, 
+				       (GTranslateFunc)dgettext_swapped,
+				       g_strdup (domain),
+				       g_free);
+}
+
+/**
+ * g_option_context_set_summary:
+ * @context: a #GOptionContext
+ * @summary: a string to be shown in <option>--help</option> output 
+ *  before the list of options, or %NULL
+ * 
+ * Adds a string to be displayed in <option>--help</option> output
+ * before the list of options. This is typically a summary of the
+ * program functionality. 
+ *
+ * Note that the summary is translated (see 
+ * g_option_context_set_translate_func(), g_option_context_set_translation_domain()).
+ *
+ * Since: 2.12
+ */
+void
+g_option_context_set_summary (GOptionContext *context,
+			      const gchar    *summary)
+{
+  g_return_if_fail (context != NULL);
+
+  g_free (context->summary);
+  context->summary = g_strdup (summary);
+}
+
+
+/**
+ * g_option_context_get_summary:
+ * @context: a #GOptionContext
+ * 
+ * Returns the summary. See g_option_context_set_summary().
+ *
+ * Returns: the summary
+ *
+ * Since: 2.12
+ */
+G_CONST_RETURN gchar *
+g_option_context_get_summary (GOptionContext *context)
+{
+  g_return_val_if_fail (context != NULL, NULL);
+
+  return context->summary;
+}
+
+/**
+ * g_option_context_set_description:
+ * @context: a #GOptionContext
+ * @description: a string to be shown in <option>--help</option> output 
+ *   after the list of options, or %NULL
+ * 
+ * Adds a string to be displayed in <option>--help</option> output
+ * after the list of options. This text often includes a bug reporting
+ * address.
+ *
+ * Note that the summary is translated (see 
+ * g_option_context_set_translate_func()).
+ *
+ * Since: 2.12
+ */
+void
+g_option_context_set_description (GOptionContext *context,
+			          const gchar    *description)
+{
+  g_return_if_fail (context != NULL);
+
+  g_free (context->description);
+  context->description = g_strdup (description);
+}
+
+
+/**
+ * g_option_context_get_description:
+ * @context: a #GOptionContext
+ * 
+ * Returns the description. See g_option_context_set_description().
+ *
+ * Returns: the description
+ *
+ * Since: 2.12
+ */
+G_CONST_RETURN gchar *
+g_option_context_get_description (GOptionContext *context)
+{
+  g_return_val_if_fail (context != NULL, NULL);
+
+  return context->description;
+}
+
+
 #define __G_OPTION_C__
 #include "galiasdef.c"
